/*
 * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 *
 * (c) Copyright 1996, 1997, 1998, 1999 Gary Henderson (gary@daniver.demon.co.uk) and
 *                                      Jerremy Koot (jkoot@snes9x.com)
 *
 * Super FX C emulator code 
 * (c) Copyright 1997, 1998, 1999 Ivar (Ivar@snes9x.com) and
 *                                Gary Henderson.
 * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
 *
 * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
 * DOS port code contains the works of other authors. See headers in
 * individual files.
 *
 * Snes9x homepage: www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */
/* Snes9x ffmpeg module was written by Bisqwit (http://iki.fi/bisqwit/)
 * Sat Sep 22 2001
 */


#include <stdio.h>

#include "record.h"

/* Get so. -structure to see the sound quality settings */
#include "soundux.h"

/* defines the snes9x pixel format */
#include "pixform.h"

#if 0

extern "C" {
#include "avformat.h"
}

/* Audio codec - hardcoded here. */
#define Audiocodec   CODEC_ID_MP2
#define Audioencoder mp2_encoder
#define Audiocodecstr "MP2"
#define AudioBPS 56000 /* per channel */

/* Video codec - hardcoded here. */
#define Videocodec   CODEC_ID_MPEG4
#define Videoencoder mpeg4_encoder
#define Videocodecstr "MPEG4"
#define MotionMethod ME_PHODS
#define VideoBPS 20000
#define VIDEOQUALITY 15

#define Filename "snes9xcap.avi"
#define Fileformat avi_format

#include <unistd.h>
#include <stdarg.h>
#include <sys/time.h>

#if 0
static void DbgWrite(const char *fmt...)
{
    char Buf[512];
    va_list ap;
    va_start(ap, fmt);
    /* stderr */
    write(2, Buf, vsprintf(Buf, fmt, ap));
    va_end(ap);
}
#else
#define DbgWrite(...)
#endif

static UINT64 getutime(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000000LL) + tv.tv_usec;
}

static int SafeWrite(char pipname, int fd, const UINT8 *buf, int count)
{
    int written = 0;
    DbgWrite("  pid %d: Writing %d bytes to %c\n", getpid(), count, pipname);
    const UINT8 *src = buf;
    while(written < count)
    {
        int ret = write(fd, src, count-written);
        //DbgWrite(" --> wrote %d to %c\n", ret, pipname);
        if(ret < 0) { written = ret; break; }
        src     += ret;
        written += ret;
    }
    
    char SBuf[256];
    sprintf(SBuf, "  pid %d: Did write %d to %c", getpid(), written, pipname);
    if(written == 4)sprintf(strchr(SBuf,0), " (value: %u)", *(unsigned *)buf);
    else if(written == 8)sprintf(strchr(SBuf, 0), " (values: %u,%u)", *(unsigned *)buf, *(unsigned *)(buf+4));
    DbgWrite("%s\n", SBuf);
    
    return written;
}
static int WaitRead(char pipname, int fd, UINT8 *buf, int count)
{
    int red = 0;
    DbgWrite("  pid %d: Reading %d bytes from %c\n", getpid(), count, pipname);
    UINT8 *dest = buf;
    while(red < count)
    {
        int ret = read(fd, dest, count-red);
        //DbgWrite(" --> %d read from %c\n", ret, pipname);
        if(ret <= 0) { red = ret; break; }
        dest += ret;
        red += ret;
    }
    char SBuf[256];
    sprintf(SBuf, "  pid %d: Did read %d from %c", getpid(), red, pipname);
    if(red == 4)sprintf(strchr(SBuf,0), " (value: %u)", *(unsigned *)buf);
    else if(red == 8)sprintf(strchr(SBuf, 0), " (values: %u,%u)", *(unsigned *)buf, *(unsigned *)(buf+4));
    DbgWrite("%s\n", SBuf);
    return red;
}

class audio
{
    AVCodec        *codec;
    AVCodecContext &c;
    
    const char *errmsg;
    
public:
    FifoBuffer fifo;
    
public:
    const char *error() const { return errmsg; }
    
    audio(AVCodecContext *cc) : c(*cc), errmsg(NULL), byleft(0)
    {
        register_avcodec(&Audioencoder);
        
        memset(&c, 0, sizeof c);
        c.codec_type  = CODEC_TYPE_AUDIO;
        c.codec_id    = Audiocodec;
        codec = avcodec_find_encoder(Audiocodec);
        if(!codec) { errmsg = "Codec "Audiocodecstr" not found"; return; }
        
        c.sample_rate = so.playback_rate;
        c.channels    = so.stereo ? 2 : 1;
        c.bit_rate    = AudioBPS*c.channels;
        c.flags |= CODEC_FLAG_HQ;
        c.quality = 1;
        
        if(avcodec_open(&c, codec) < 0) { errmsg = "Could not open "Audiocodecstr" codec"; return; }
        
        fifo_init(&fifo, c.frame_size * 2 * c.channels * 2); /* last *2=to be safe */
    }
    
    AVCodecContext *getcontext() { return &c; }    
    
    void addsamples(const UINT8 *data, unsigned bytecount)
    {
        fifo_write(&fifo, const_cast<UINT8 *> (data), bytecount,  &fifo.wptr);
    }
        
    ~audio()
    {
        avcodec_close(&c);
    }
private:
    unsigned byleft;
   
public:
    void ReadPipe(int pipfd)
    {
        if(!byleft)
        {
            WaitRead('A', pipfd, (UINT8 *)&byleft, sizeof byleft);
            return;
        }
        UINT8 buf[byleft];
        int got = read(pipfd, buf, byleft); /* don't block */
        DbgWrite("             Audio: read %d bytes\n", got);
        if(got <= 0)return;
        
        RecordAudio(buf, got);
        byleft -= got;
    }
};


class video
{
    AVCodec        *codec;
    AVCodecContext &c;
    AVPicture      pic, pic420;
    
    const char *errmsg;
    
    UINT8 *RGBbuf;
    UINT8 *P420;
    
public:
    const char *error() const { return errmsg; }
    
    video(AVCodecContext *cc) : c(*cc), errmsg(NULL), byleft(0)
    {
        register_avcodec(&Videoencoder);

        memset(&c, 0, sizeof c);
        c.codec_type  = CODEC_TYPE_VIDEO;
        c.codec_id    = Videocodec;

        codec = avcodec_find_encoder(Videocodec);
        if(!codec) { errmsg = "Codec "Videocodecstr" not found"; return; }
        
        c.frame_rate  = 13 * FRAME_RATE_BASE;
        c.bit_rate    = VideoBPS;
        c.width       = 256;
        c.height      = 224;
        c.gop_size    = 50;     // key frame step
        c.pix_fmt     = PIX_FMT_YUV420P;
        
        if(VIDEOQUALITY)
        {
	        c.flags |= CODEC_FLAG_QSCALE;
	        c.quality = VIDEOQUALITY;
	    }
        
        if(avcodec_open(&c, codec) < 0) { errmsg = "Could not open "Videocodecstr" codec"; return; }
        
        RGBbuf = new UINT8[c.width * c.height * 3];
        P420   = new UINT8[(c.width * c.height * 3) / 2];
    }
    
    AVCodecContext *getcontext() { return &c; }    
    
    AVPicture *addframe(const uint8 *vidbuf, unsigned width, unsigned height)
    {
        memset(&pic, 0, sizeof(pic));        
        avpicture_fill(&pic, RGBbuf, PIX_FMT_RGB24, c.width, c.height);
        
        // FIXME: This would require dynamic
        // reallocation of Rbuf[], Gbuf[] and Bbuf[].
        
        width = c.width;  // c.width       = width;
        height= c.height; // c.height      = height;
        
        unsigned pixcount = width*height;
        
        register const uint16 *pixp = (const uint16 *)vidbuf;
        
        UINT8 *dest = RGBbuf;
        for(unsigned p=0; p<pixcount; ++p)
        {
            register uint16 pixel = *pixp++;
            *dest++ = (((pixel/RED_LOW_BIT_MASK  ) & MAX_RED  ) * (0x100 / (MAX_RED  +1)));
            *dest++ = (((pixel/GREEN_LOW_BIT_MASK) & MAX_GREEN) * (0x100 / (MAX_GREEN+1)));
            *dest++ = (((pixel/BLUE_LOW_BIT_MASK ) & MAX_BLUE ) * (0x100 / (MAX_BLUE +1)));
        }
        
        if(c.pix_fmt == PIX_FMT_RGB24)
            return &pic;
        
        avpicture_fill(&pic420, P420, c.pix_fmt, c.width, c.height);
        img_convert(&pic420, c.pix_fmt, &pic, PIX_FMT_RGB24, c.width, c.height);
        return &pic420;
    }
    
    ~video()
    {
        avcodec_close(&c);
        delete[] RGBbuf;
        delete[] P420;
    }
    
private:
    unsigned byhei, bywi;
    unsigned bygot, byleft;
    UINT8 *bybuf;
public:
    void ReadPipe(int pipfd)
    {
        if(!byleft)
        {
            unsigned arr[2];
            WaitRead('V', pipfd, (UINT8 *)&arr, sizeof arr);
            bywi = arr[0];
            byhei= arr[1];
            byleft = bywi * byhei * 2;
            bybuf = new UINT8[byleft];
            bygot = 0;
            DbgWrite("             Video: Want %u bytes (%u * %u)\n", byleft, bywi, byhei);
            return;
        }
        int got = read(pipfd, bybuf + bygot, byleft);
        //DbgWrite("             Video: read %d bytes\n", got);
        if(got <= 0)return;
        
        bygot += got;
        byleft -= got;
        
        if(!byleft)
        {
            DbgWrite("             Video: That was all\n");
            RecordFrame(bybuf, bywi, byhei);
            delete[] bybuf;
        }
    }
};


class muxer
{
    AVFormatContext oc;

    const char *errmsg;
    
    audio *rec_audio;
    video *rec_video;

    unsigned frameno, frame;
    UINT64 starttime;
    UINT64 pausetime;
    
public:
    const char *error() const { return errmsg; }
    
    muxer() : errmsg(NULL), rec_audio(NULL), rec_video(NULL)
    {
        avcodec_init();
        register_avformat(&Fileformat);
        register_protocol(&file_protocol);

        motion_estimation_method = MotionMethod;
    
        memset(&oc, 0, sizeof(oc));
        
        AVFormat *format = guess_format(NULL, Filename, NULL);
        if(!format)format = &Fileformat;
        
        oc.format = format;
        
        /* FIXME */
        strcpy(oc.title,     "snes9x capture");
        strcpy(oc.author,    "Joel Yliluoma");
        strcpy(oc.copyright, "Copyright (C) 1992, 2001 Bisqwit (http://iki.fi/bisqwit/)");
        strcpy(oc.comment,   "Game: Chrono Trigger");
        
	    strcpy(oc.filename, Filename);
	    AVStream *st;
	    
	    st = new AVStream;
	    memset(st, 0, sizeof(*st));
        rec_video = new video(&st->codec);
        oc.streams[oc.nb_streams++] = st;
        if(rec_video->error())
        {
            errmsg = rec_video->error();
            oc.format = NULL;
            return;
        }
        
        st = new AVStream;
	    memset(st, 0, sizeof(*st));
        rec_audio = new audio(&st->codec);
        oc.streams[oc.nb_streams++] = st;
        if(rec_audio->error())
        {
            errmsg = rec_audio->error();
            oc.format = NULL;
            return;
        }

        dump_format(&oc, 1, oc.filename, 1);
        fflush(stderr);

        if(url_fopen(&oc.pb, oc.filename, URL_WRONLY) < 0)
        {
            errmsg = Filename" can't be written";
            oc.format = NULL;
            return;
        }
        
        oc.format->write_header(&oc);
        
        frameno   = 0;
        starttime = getutime();
        frame = 0;
    }
    ~muxer()
    {
        if(oc.format)
        {
            oc.format->write_trailer(&oc);
            url_fclose(&oc.pb);
        }
        
        delete rec_audio;
        delete rec_video;
    }
    
    void startpause()
    {
    	pausetime = getutime();
    }
    void endpause()
    {
    	starttime += (getutime() - pausetime);
    }
    
    void addframe(const uint8 *vidbuf, unsigned width, unsigned height)
    {
        if(!rec_video)return;
        
        AVPicture *pic = rec_video->addframe(vidbuf, width, height);
        if(!pic)return;
        
        UINT8 buffer[1048576];
        int ret = 0;
        
        for(;;)
        {
            UINT64 havetime = getutime() - starttime;
            UINT64 timeperframe = (1000000LL * FRAME_RATE_BASE / rec_video->getcontext()->frame_rate);
            UINT64 wanttime = timeperframe * frameno;
            
            DbgWrite("    havetime=%Lu, timeperframe=%Lu, wanttime=%Lu\n", havetime, timeperframe, wanttime);
            
            if(!timeperframe)break;
            
            if(wanttime > havetime+timeperframe)break;
            
            if(!ret) /* encode a frame only once */
            {
                ret = avcodec_encode_video(rec_video->getcontext(), buffer, sizeof buffer, pic);
            }
            
            oc.format->write_packet(&oc, 0 /* stream no */, buffer, ret);
            ++frameno;
            DbgWrite("    frame%u: wrote video frame %u (%u bytes)\n", frame++, frameno, ret);
        }
    }
    
    void addsamples(const UINT8 *data, unsigned bytecount)
    {
        if(!rec_audio)return;
        
        rec_audio->addsamples(data, bytecount);
        
        unsigned framesize = 
          rec_audio->getcontext()->frame_size * 2
        * rec_audio->getcontext()->channels;
        
        UINT8 buf[framesize];
        
        DbgWrite("want %u bytes of audio, %u available...\n",
            framesize,
            fifo_size(&rec_audio->fifo, rec_audio->fifo.rptr)
        );

        while(fifo_read(&rec_audio->fifo, buf, framesize,
                        &rec_audio->fifo.rptr) == 0)
        {
            UINT8 obuf[8192];
            int ret = avcodec_encode_audio(
                    rec_audio->getcontext(),
                    obuf, sizeof obuf,
                    (const short *)buf);
            oc.format->write_packet(&oc, 1 /* stream no */, obuf, ret);
            DbgWrite("frame%u: wrote audio packet (%u bytes)\n", frame++, ret);
        }
    }
    void ReadVPipe(int fd) { if(rec_video) rec_video->ReadPipe(fd); }
    void ReadAPipe(int fd) { if(rec_audio) rec_audio->ReadPipe(fd); }
};

#include <signal.h>

static int encoderpid = 0;
static int audiofd, videofd;

static volatile int terminate = 0;
static muxer *mux = NULL;

#include <sys/poll.h>

static void RecordKiller(int)
{
    ++terminate;
}

#include <sys/socket.h>

void RecordInit()
{
    if(!encoderpid)
    {
        int apip[2], vpip[2];
        
        /* We use socketpair here instead of pipe, because sockets come
         * with greater buffers. Pipes have only 4096 byte buffers.
         * This improves performance, as the two processes don't
         * need to wait each others all the time.
         */
        
        /* We can't use the same pipe for both video and audio, because
         * audio output may interrupt video output (either thread or sigalarm).
         */
        
        socketpair(AF_UNIX, SOCK_STREAM, 0, apip);   //pipe(apip);
        socketpair(AF_UNIX, SOCK_STREAM, 0, vpip);   //pipe(vpip);
        
        encoderpid = fork();
        if(!encoderpid)
        {
            DbgWrite(">RecordInit in pid %d\n", getpid());
            
            signal(SIGTERM, RecordKiller);
            signal(SIGPIPE, RecordKiller);
            signal(SIGHUP,  RecordKiller);
            signal(SIGALRM, SIG_IGN);
    
            mux = new muxer();
            if(mux->error())
            {
                DbgWrite("%s\n", mux->error());
                delete mux;
                mux = NULL;
                _exit(-1);
            }
            
            DbgWrite("<RecordInit in pid %d\n", getpid());
            
            close(vpip[1]);
            close(apip[1]);
            
            while(!terminate)
            {
                struct pollfd fds[2] =
                { { vpip[0], POLLIN, 0 },
                  { apip[0], POLLIN, 0 }
                };
                
                if(!poll(fds, 2, 60))continue;
                
                if(fds[0].revents & (POLLERR|POLLHUP))break;
                if(fds[1].revents & (POLLERR|POLLHUP))break;
                
                if(fds[0].revents & (POLLIN|POLLPRI))mux->ReadVPipe(vpip[0]);
                if(fds[1].revents & (POLLIN|POLLPRI))mux->ReadAPipe(apip[0]);
            }
            DbgWrite("recording ending\n");
            RecordDone();
            _exit(0);
        }
        
        close(apip[0]); audiofd = apip[1];
        close(vpip[0]); videofd = vpip[1];
    }
}

void RecordDone()
{
    DbgWrite(">RecordDone in pid %d\n", getpid());
    if(encoderpid)
    {
        kill(encoderpid, SIGHUP);
        close(audiofd);
        close(videofd);
        encoderpid = 0;
    }
    else
    {
        delete mux;
        mux = NULL;
    }
    DbgWrite("<RecordDone in pid %d\n", getpid());
}

void RecordFrame(const uint8 *vidbuf, unsigned width, unsigned height)
{
    if(encoderpid)
    {
        DbgWrite("---WRITING V---\n");
        unsigned arr[2] = {width, height};
        SafeWrite('V', videofd, (const UINT8 *)&arr, sizeof arr);
        SafeWrite('V', videofd, (const UINT8 *)vidbuf, width*height*2);
        DbgWrite("---WROTE V---\n");
    }
    if(mux)
        mux->addframe(vidbuf, width, height);
}

void RecordAudio(const uint8 *buf, unsigned bytecount)
{
    if(encoderpid)
    {
        DbgWrite("---WRITING A---\n");
        SafeWrite('A', audiofd, (const UINT8 *)&bytecount, sizeof bytecount);
        SafeWrite('A', audiofd, (const UINT8 *)buf, bytecount);
        DbgWrite("---WROTE A---\n");
    }
    if(mux)
        mux->addsamples(buf, bytecount);
}

void RecordPause()
{
	if(mux) mux->startpause();
}

void RecordUnpause()
{
	if(mux) mux->endpause();
}
#else

void RecordInit() {}
void RecordDone() {}
void RecordFrame(const uint8 *vidbuf, unsigned width, unsigned height) {}
void RecordAudio(const uint8 *buf, unsigned bytecount) {}
void RecordPause() {}
void RecordUnpause() {}

#endif
